@@ -1,2 +1,2 @@
-from .data_loader import load_nhefs, load_acic16
+from .data_loader import load_nhefs, load_nhefs_survival, load_acic16
 from ..simulation.CausalSimulator3 import CausalSimulator3 as CausalSimulator, generate_random_topology
