@@ -1,6 +1,10 @@
-from .doubly_robust import DoublyRobustIpFeature, DoublyRobustJoffe, DoublyRobustVanilla
+from .doubly_robust import AIPW, PropensityFeatureStandardization, WeightedStandardization
 from .ipw import IPW
 from .overlap_weights import OverlapWeights
 from .standardization import Standardization, StratifiedStandardization
 from .marginal_outcome import MarginalOutcomeEstimator
-from .matching import Matching, PropensityMatching
\ No newline at end of file
+from .matching import Matching, PropensityMatching
+from .rlearner import RLearner
+from .xlearner import XLearner
+from .tmle import TMLE
+
