@@ -26,13 +26,15 @@ import abc
 import warnings
 
 import pandas as pd
+import numpy as np
 
 from .base_estimator import IndividualOutcomeEstimator
-from .base_weight import WeightEstimator
+from .base_weight import WeightEstimator, PropensityEstimator
 from ..utils import general_tools as g_tools
+from ..utils.stat_utils import robust_lookup
 
 
-class DoublyRobust(IndividualOutcomeEstimator):
+class BaseDoublyRobust(IndividualOutcomeEstimator):
     """
     Abstract class defining the interface and general initialization of specific doubly-robust methods.
     """
@@ -52,7 +54,7 @@ class DoublyRobust(IndividualOutcomeEstimator):
                                        If None - all covariates passed will be used.
                                        Either list of column names or boolean mask.
         """
-        super(DoublyRobust, self).__init__(lambda **x: None)  # Dummy initialization
+        super(BaseDoublyRobust, self).__init__(lambda **x: None)  # Dummy initialization
         delattr(self, "learner")  # To remove the learner attribute a IndividualOutcomeEstimator has
         self.outcome_model = outcome_model
         self.weight_model = weight_model
@@ -98,28 +100,64 @@ class DoublyRobust(IndividualOutcomeEstimator):
         return repr_string
 
 
-class DoublyRobustVanilla(DoublyRobust):
-    """
-    Given the measured outcome Y, the assignment Y, and the coefficients X calculate a doubly-robust estimator
-    of the effect of treatment
+class AIPW(BaseDoublyRobust):
+    def __init__(self, outcome_model, weight_model,
+                 outcome_covariates=None, weight_covariates=None,
+                 overlap_weighting=False):
+        """
+        Calculates a doubly-robust estimate of the treatment effect by performing
+        potential-outcome prediction (`outcome_model`) and then correcting its
+        prediction-residuals using re-weighting from a treatment model (`weight_model`, like IPW).
 
-    Let e(X) be the estimated propensity score and m(X, A) is the estimated effect by an estimator,
-    then the individual estimates are:
+        It has two flavors, which slightly change the weighting of the outcome model in the correction term.
+        Let e(X) be the estimated propensity score and m(X, A) is the estimated effect by an estimator,
+        then the individual estimates are:
 
-    | Y + (A-e(X))*(Y-m(X,1)) / e(X) if A==1, and
-    | Y + (e(X)-A)*(Y-m(X,0)) / (1-e(X)) if A==0
+        | m(X,1) + A*(Y-m(X,1))/e(X), and
+        | m(X,0) + (1-A)*(Y-m(X,0))/(1-e(X))
 
-    These expressions show that when e(X) is an unbiased estimator of A, or when m is an unbiased estimator of Y
-    then the resulting estimator is unbiased. Note that the term for A==0 is derived from (1-A)-(1-e(X))
+        Which are basically add IP-weighted residuals from the observed predictions.
+        As described in Kang and Schafer (2007) section 3.1 and Robins, Rotnitzky, and Zhao (1994).
 
-    Another way of writing these equation is by "correcting" the individual prediction rather than the individual
-    outcome:
+        The additional flavor when `overlap_weighting=True` is from Glynn and Quinn (2010),
+        adds weighting by the propensity-of-the-other-class to the outcome model,
+        so extreme example (with poor covariate overlap) will contribute less to the correction
+        (i.e. rely less on their prediction value that might extrapolate too much).
+        This is a similar notion used in Overlap Weights model (hence the argument name)
 
-    | m(X,1) + A*(Y-m(X,1))/e(X), and
-    | m(X,0) + (1-A)*(Y-m(X,0))/(1-e(X))
-    """
+        | A * [Y - (1-e(X))m(X,1)]/e(X) + (1-A) * m(X,1), and
+        | (1-A) * [Y - e(X)m(X,0)]/(1-e(X)) + A * m(X,0)
+
+        Args:
+            outcome_model(IndividualOutcomeEstimator): A causal model that estimate on individuals level
+                                                      (e.g. Standardization).
+            weight_model (WeightEstimator | PropensityEstimator): A causal model for weighting individuals (e.g. IPW).
+                If `overlap_weighting=True` then must be a `PropensityEstimator` model.
+            outcome_covariates (array): Covariates to use for outcome model.
+                If None - all covariates passed will be used. Either list of column names or boolean mask.
+            weight_covariates (array): Covariates to use for weight model.
+                If None - all covariates passed will be used. Either list of column names or boolean mask.
+            overlap_weighting (bool): Whether to tweak the outcome-model correction-term to rely less on
+                data-points with poor covariate overlap (extreme propensity).
+                if `True`, requires `weight_model` to be an instance of `PropensityEstimator`.
+
+        References:
+            * Kang and Schafer, 2007, (https://dx.doi.org/10.1214/07-STS227)
+            * Kang and Schafer attribute the original method to Cassel, Särndal and Wretman.
+            * Glynn and Quinn, 2010, https://doi.org/10.1093/pan/mpp036
+            * Robins, Rotnitzky, and Zhao, 1994, https://doi.org/10.1080/01621459.1994.10476818
+        """
+        super().__init__(outcome_model, weight_model,
+                         outcome_covariates, weight_covariates)
+        self.overlap_weighting = overlap_weighting
 
     def fit(self, X, a, y, refit_weight_model=True, **kwargs):
+        if self.overlap_weighting and a.nunique() != 2:
+            raise AssertionError(
+                f"`overlap_weights=True` version can only be used with binary treatment."
+                f"Instead, treatment values are {set(a)}. Try setting it to `False`."
+            )
+
         X_outcome, X_weight = self._prepare_data(X, a)
         weight_model_is_not_fitted = not g_tools.check_learner_is_fitted(self.weight_model.learner)
 
@@ -187,21 +225,30 @@ class DoublyRobustVanilla(DoublyRobust):
                           treatment value in the corresponding key.
         """
         X_outcome, X_weight = self._prepare_data(X, a)
-        individual_cf = self.estimate_individual_outcome(X_outcome, a, treatment_values, predict_proba)
         weights = self.weight_model.compute_weights(X_weight, a)
-        # Correct individual-estimation for later averaging:
+        individual_cf = self.estimate_individual_outcome(X_outcome, a, treatment_values, predict_proba)
+        is_predict_proba_classification_result = isinstance(individual_cf.columns, pd.MultiIndex)
+        if is_predict_proba_classification_result:
+            # Classification `outcome_model` with `predict_proba=True` returns a MultiIndex of treatments over outcomes.
+            # Extract the prediction for the maximal outcome class (probably class `1` in binary classification):
+            outcome_values = individual_cf.columns.get_level_values(level=-1)
+            individual_cf = individual_cf.xs(
+                outcome_values.max(), axis="columns", level=-1, drop_level=True,
+            )
+        factual_prediction = robust_lookup(individual_cf, a)
+
+        if self.overlap_weighting:
+            propensities = self.weight_model.compute_propensity_matrix(X_weight)
+            reversed_propensities = robust_lookup(propensities, 1 - a)  # take propensities of opposite treatment group
+            factual_prediction *= reversed_propensities
+            corrected_outcome = (y - factual_prediction) * weights
+        else:
+            outcome_correction = (y - factual_prediction) * weights
+            corrected_outcome = factual_prediction + outcome_correction
+
         for treatment_value in treatment_values:
-            is_treated = a == treatment_value
-            y_cur = y.loc[is_treated]
-            y_pred_cur = individual_cf.loc[is_treated, treatment_value]
-            w_cur = weights.loc[is_treated]
-            # This is a the same as (y_cur - y_pred_cur) * w_cur, but compatible with y_pred being both Series (due to
-            # predict_proba=False) and DataFrame (predict_proba=True).
-            correction = y_pred_cur.mul(-1.0).add(y_cur, axis="index").mul(w_cur, axis="index")
-            individual_cf.loc[is_treated, treatment_value] += correction
-        # NOTE: that we do not use the corrected-counter-factual-outcome vector when we return our individual effect.
-        # This is because this correction "contaminates" the model prediction with the actual outcome.
-        # Causing the model results to be inconsistent with the model itself.
+            individual_cf.loc[a == treatment_value, treatment_value] = corrected_outcome.loc[a == treatment_value]
+
         return individual_cf
 
     def estimate_population_outcome(self, X, a, y=None, treatment_values=None, predict_proba=None, agg_func="mean"):
@@ -247,15 +294,56 @@ class DoublyRobustVanilla(DoublyRobust):
                           "not corrected for population effect.\n"
                           "In case you want individual effect use agg='individual', or in case you want population"
                           "effect use the estimate_population_effect() output as your input to this function.")
-        effect = super(DoublyRobustVanilla, self).estimate_effect(outcome1, outcome2, agg, effect_types)
+        effect = super().estimate_effect(outcome1, outcome2, agg, effect_types)
         return effect
 
 
-class DoublyRobustIpFeature(DoublyRobust):
-    """
-    A doubly-robust estimator of the effect of treatment.
-    This model adds the weighting (inverse probability weighting) as feature to the model.
-    """
+class PropensityFeatureStandardization(BaseDoublyRobust):
+    def __init__(self, outcome_model, weight_model,
+                 outcome_covariates=None, weight_covariates=None,
+                 feature_type="weight_vector"):
+        """
+        A doubly-robust estimator of the effect of treatment.
+        This model adds the weighting (inverse probability weighting)
+        as additional feature to the outcome model.
+
+        References:
+            * Bang and Robins, https://doi.org/10.1111/j.1541-0420.2005.00377.x
+            * Kang and Schafer, section 3.3, https://dx.doi.org/10.1214/07-STS227
+
+        Args:
+            outcome_model(IndividualOutcomeEstimator): A causal model that estimate on individuals level
+            weight_model (WeightEstimator | PropensityEstimator): A causal model for weighting individuals (e.g. IPW).
+            outcome_covariates (array): Covariates to use for outcome model.
+                If None - all covariates passed will be used. Either list of column names or boolean mask.
+            weight_covariates (array): Covariates to use for weight model.
+                If None - all covariates passed will be used. Either list of column names or boolean mask.
+            feature_type (str): the type of covariate to add. One of the following options:
+                *  "weight_vector": uses a signed weight vector. Only defined for binary treatment.
+                   For example, if `weight_model` is IPW then: 1/Pr[A=a_i|X] for each sample `i`.
+                   As described in Bang and Robins (2005).
+                * "signed_weight_vector": as `'weight_vector'`, but negates the weights of the control group.
+                  For example, if `weight_model` is IPW then: 1/Pr[A|X] for treated and 1/Pr[A|X] for controls.
+                  As described in the correction for Bang and Robins (2008)
+                * "weight_matrix": uses the entire weight matrix.
+                   For example, if `weight_model` is IPW then: 1/Pr[A_i=a|X_i=x],
+                                for all treatment values `a` and for every sample `i`.
+                * "masked_weighted_matrix": uses the entire weight matrix, but masks it with a dummy-encoding
+                  of the treatment assignment.
+                  For example, if weight_model` is IPW then: 1/Pr[A=a_i|X=x_i] and 0 for all other `a≠a_i` columns.
+                  As described in Bang and Robins (2005).
+                * "propensity_vector": uses the probabilities for being in treatment group: Pr[A=1|X].
+                                       Better defined for binary treatment.
+                                       Equivalent to Scharfstein, Rotnitzky, and Robins (1999) that use its inverse.
+                * "logit_propensity_vector": uses logit transformation of the propensity to treat Pr[A=1|X].
+                                             As described in Kang and Schafer (2007)
+                * "propensity_matrix": uses the probabilities for all treatment options,
+                    Pr[A_i=a|X_i=x] for all treatment values `a` and samples `i`.
+        """
+        super().__init__(outcome_model, weight_model,
+                         outcome_covariates, weight_covariates)
+        self.feature_type = feature_type
+
 
     def estimate_individual_outcome(self, X, a, treatment_values=None, predict_proba=None):
         X_augmented = self._augment_outcome_model_data(X, a)
@@ -276,12 +364,8 @@ class DoublyRobustIpFeature(DoublyRobust):
                           matrix (W | X).
         """
         X_outcome, X_weight = self._prepare_data(X, a)
-        try:
-            weights_feature = self.weight_model.compute_weight_matrix(X_weight, a)
-            weights_feature = weights_feature.add_prefix("ipf_")
-        except NotImplementedError:
-            weights_feature = self.weight_model.compute_weights(X_weight, a)
-            weights_feature = weights_feature.rename("ipf")
+        feature_func = self._get_feature_function(self.feature_type)
+        weights_feature = feature_func(X_weight, a)
         # Let standardization deal with incorporating treatment assignment (a) into the data:
         X_augmented = pd.concat([weights_feature, X_outcome], join="outer", axis="columns")
         return X_augmented
@@ -300,12 +384,74 @@ class DoublyRobustIpFeature(DoublyRobust):
         self.outcome_model.fit(X=X_augmented, y=y, a=a)
         return self
 
+    def _get_feature_function(self, function_name):
 
-class DoublyRobustJoffe(DoublyRobust):
+        def weight_vector(X, a):
+            w = self.weight_model.compute_weights(X, a)
+            w = w.rename("ipf")
+            return w
+
+        def signed_weight_vector(X, a):
+            """Bang and Robins, 2005 / 2008"""
+            if a.nunique() != 2:
+                raise AssertionError(
+                    f"`feature_type` 'weight_vector' can only be used with binary treatment."
+                    f"Instead, treatment values are {set(a)}."
+                )
+            w = weight_vector(X, a)
+            w[a == 0] *= -1
+            return w
+
+        def weight_matrix(X, a):
+            W = self.weight_model.compute_weight_matrix(X, a)
+            W = W.add_prefix("ipf_")
+            return W
+
+        def masked_weight_matrix(X, a):
+            """Bang and Robins, 2005"""
+            W = weight_matrix(X, a)
+            A = pd.get_dummies(a)
+            A = A.add_prefix("ipf_")  # To match naming of `W`
+            W_masked = W * A
+            return W_masked
+
+        def propensity_vector(X, a):
+            p = self.weight_model.compute_propensity(X, a)
+            p = p.rename("propensity")
+            return p
+
+        def logit_propensity_vector(X, a, safe=True):
+            p = propensity_vector(X, a)
+            if safe:
+                epsilon = np.finfo(float).eps
+                p = np.clip(p, epsilon, 1 - epsilon)
+            return np.log(p / (1 - p))
+
+        def propensity_matrix(X, a):
+            P = self.weight_model.compute_propensity_matrix(X)
+            # P = P.iloc[:, 1:]  # Drop first column
+            P = P.add_prefix("propensity_")
+            return P
+
+        feature_functions = {
+            "weight_vector": weight_vector,
+            "signed_weight_vector": signed_weight_vector,
+            "weight_matrix": weight_matrix,
+            "masked_weight_matrix": masked_weight_matrix,
+            "propensity_vector": propensity_vector,
+            "logit_propensity_vector": logit_propensity_vector,
+            "propensity_matrix": propensity_matrix,
+        }
+        return feature_functions[function_name]
+
+
+class WeightedStandardization(BaseDoublyRobust):
     """
-    A doubly-robust estimator of the effect of treatment.
-    This model uses the weights from the weight-model (e.g. inverse probability weighting) as individual weights for
-    fitting the outcome model.
+    This model uses the weights from the weight-model (e.g. inverse probability weighting)
+    as individual weights for fitting the outcome model.
+
+    References:
+        * Kang and Schafer, section 3.2, https://dx.doi.org/10.1214/07-STS227
     """
 
     def estimate_individual_outcome(self, X, a, treatment_values=None, predict_proba=None):
